############################################
# Purpose:
# This script is indented to identify which variables are used in calls to external functions or routines.
#
# Inputs:
#
#    -s, --source : Path to the preprocessed sources.
#    -v, --vault : Path to the vault.
#    -o, --output : Output filename for the list of variables used in external calls.
#
############################################
#
from AutoRPE.UtilsRPE.CodeTools import expand_sources
from AutoRPE.UtilsRPE.Classes import Current_block, ImplementRPE, Routine, Interface
from AutoRPE.UtilsRPE.Functions import load_vault, calls_to_routines, in_which_block, \
    find_called_routine, find_call_arguments, retrieve_variable, save_variable_list
from AutoRPE.UtilsRPE.Error import ProcedureNotFound
from os.path import basename

real_types = ["single", "double"]


def get_command_line_arguments():
    """
    Parse and check command line options
    """
    from optparse import OptionParser
    from os.path import isfile, isdir
    # Parse and assert command line options
    parser = OptionParser()
    parser.add_option("-s", "--source", dest="source_path",
                      help="Path to the preprocessed source.", default=None)
    parser.add_option("-v", "--vault", dest="vault_path",
                      help="Path to the vault.", default=None, metavar="FILE")
    parser.add_option("-o", "--output", dest="output",
                      help="Output filename for the list of variables used in external calls.",
                      default=None, metavar="FILE")

    (options, args) = parser.parse_args()

    source_path = options.source_path
    vault_path = options.vault_path
    output_filename = options.output
    # Some checks for the arguments provided.
    try:
        assert source_path is not None, "Argument -s/--source required!"
        assert vault_path is not None, "Argument -v/--vault required!"
        assert isdir(source_path), "Argument -s/--source  %s is not a valid directory." % source_path
        assert isfile(vault_path), "Argument -v/--vault %s is not a file." % vault_path
    except AssertionError as error:
        parser.print_help()
        print("\n\n")
        print(error)
        exit(1)

    return source_path, vault_path, output_filename


source_path, vault_path, output_filename = get_command_line_arguments()

# TODO: Some NEMO specific exceptions that shouldn't be necessary.
exceptions = ["mpi_", "xios_", "agrif_", "cice", "ice_", "wheneq", "trc_", "flush"]

# Load vault
vault = load_vault(vault_path)

# Find all sources inside source_path
files = expand_sources(source_path)

# Initialize empty lists for external routines and variables used in external routines.
external_routines_found = []
variables = []

# Loop over source files
for filepath in files:
    module_name = basename(filepath).split(".")[0]
    module = vault[module_name]
    if module is None:
        continue
    # Use existing class to process the file
    file_info = ImplementRPE(filepath, vault)
    file_info.pre_process()

    # Initialize current_block
    current_block = Current_block(module.main())
    # Loop over lines in the file
    for index, line in enumerate(file_info.lines):
        # Find calls to routines
        try:
            routine_calls = calls_to_routines(line)
        except ValueError:
            continue
        # In case there are routine calls ...
        if routine_calls:
            for call in routine_calls:
                is_external = False
                called_routine = find_called_routine(call)

                # Find the procedure called (object)
                try:
                    procedure = vault.get_procedure(called_routine)
                except ProcedureNotFound:
                    procedure = None

                # In case it corresponds to a routine that exists in the vault, continue, otherwise mark it as external.
                if isinstance(procedure, Routine):
                    continue
                elif isinstance(procedure, Interface):
                    continue
                else:
                    is_external = True

                # In case a call to an external routine is found ...
                if is_external:
                    external_routines_found.append(called_routine)
                    # Looking if the call its a call to one of our predefined exceptions.
                    is_exception = any([called_routine.count(exception) > 0 for exception in exceptions])
                    if is_exception:
                        continue
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        # If there's no argument, continue
                        continue
                    else:
                        # In case we have an external call with arguments, loop over the arguments
                        for argument in called_arguments:
                            try:
                                variable = retrieve_variable(argument, routine=current_block.routine, vault=vault)
                            except AssertionError:
                                continue
                            except ValueError:
                                continue
                            # If a variable is identified and the variable is a real, add its id to the list.
                            if variable is not None:
                                if variable.type in real_types:
                                    variables.append(variable.id)
        else:
            current_block = in_which_block(line, current_block, vault=vault)

# Avoid duplicates in the list
variables = list(set(variables))

# Save the list of variables in the provided output filename.
save_variable_list(variables, output_filename, vault)
