import pickle
import re
import warnings
from os.path import isfile, join, dirname, abspath
from AutoRPE.UtilsRPE.Error import PointerTarged, ProcedureNotFound, MemberNotFound
import AutoRPE.UtilsRPE.precision_module as precision_module

# We have a problem with the working precision, which can be different in different moments
# i.e. double when implementing emulator, single when modifying code for final implementation
single = "single"
double = "double"
# Using the precision module to set the working precision
working = precision_module.working_precision
real_types = [single, double]

sp = single
dp = double
wp = working
lookup_table = {'WP': wp, 'default': wp, '8': dp, 'dp': dp, 'sp': sp, 'fbsp': sp, 'wp': wp, 'KIND=dp': dp,
                "KIND=wp": wp, 'kind=fbsp': sp, "kind=sp": sp, "kind=wp": wp, "KIND=fbdp": dp,
                "KIND=fbsp": sp, "KIND=8": dp}

pattern_real_with_prec = r"^[0-9\.]*_\w+$"

# Pattern for numbers in exponential form, or numbers with double precision exponent
pattern_exp_DPE_notation = r'\s*\d+\.?\d*[e|d][+|-]*\d+[\w]*'
vec = r"\w+\s*\(\s*[\w,:\d\s+\-\*=]+\s*\)"
pattern_vec = vec + r"[%\w]*"
pattern_struct = vec + r"[%\w]+\(\s*[\w,:\d]+\s*\)"
pattern_array = r"\(/\s*[\w\s,%.()\+\-=*]*\s*/\)"

# Hardcoded list, this should be solved
list_of_intrinsics = ["max", "maxloc", "maxval", "min", "minloc", "minval", "epsilon",
                      "isnan", "count", "abs", "mod", "reshape",
                      "atanh", "sqrt", "log",
                      "sum",
                      "erfc",
                      "dsqrt", "cmplx", "transfer"]

list_of_key = ["dim", "mask", "kind", "back"]


# Functions
def load_variable_list(filename):
    # Load list of variables
    from numpy import loadtxt
    variables = loadtxt(filename, dtype=int).tolist()
    if not isinstance(variables, list):
        variables = [variables]
    return variables


def save_variable_list(list_of_variables, filename, vault):
    # Save list of variables into a txt file
    # Sort and keep only uniq values
    list_of_variables = list(sorted(set(list_of_variables)))
    line_format = "%5i    # %20s %25s %20s\n"
    with open(filename, "w") as output_file:
        # Add header
        output_file.write("#%5s     %20s %25s %20s\n%s\n" % ("ID", "variable", "routine", "module", 100 * "#"))
        # Add variables
        for var_id in list_of_variables:
            v = vault.get_variable_by_id(var_id)
            output_file.write(line_format % (v.id, v.name, v.routine.name, v.routine.module.name))


def split_semicolon(string):
    # Trying to use a rarely used character ({) for the temporal replacement
    if string.count(";"):
        # In the case where there's quotes in the text, we should ignore the exclamations inside
        if not string.count("'") and not string.count("\""):
            return string.split(";")
        else:
            # Looking for
            scan_quotation = [string[:i].count("'") for i, c in enumerate(string)]
            scan_quotation2 = [string[:i].count("\"") for i, c in enumerate(string)]
            splitted_text = [c for c in string]
            splitted_text = ["{" if c == ";" and (scan_quotation[i] % 2 == 1 or scan_quotation2[i] % 2 == 1) else c for
                             i, c in
                             enumerate(splitted_text)]
            string = "".join(splitted_text)
            return string.split(";")

    raise AssertionError("string must contain semi-colon")


def strip_call_arguments(_line):
    # Description: Given a string with a comma separated arguments,
    # extract the list of arguments getting rid of the parenthesis
    # Input: String with subroutine declaration
    # Output: List with arguments
    op = [i for i, x in enumerate(_line) if x == "("]
    cp = [i for i, x in enumerate(_line) if x == ")"]
    _line = _line[op[0] + 1:cp[-1]]
    op = [i for i, x in enumerate(_line) if x == "("]
    cp = [i for i, x in enumerate(_line) if x == ")"]
    commas = [i for i, x in enumerate(_line) if x == ","]
    splitted_line = [char for char in _line]
    for n in commas:
        if len([x for x in op if x < n]) == len([x for x in cp if x < n]):
            splitted_line[n] = "#"
    _line = "".join(splitted_line)
    return [_x.strip() for _x in _line.split("#")]


def add_val_to_argument(string, _field):
    num_arguments, intervals = find_argument_intervals(string)
    for i in [_field]:
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        if intermediate_segment.strip()[:2] == "(/":
            intermediate_segment = intermediate_segment.strip()[2:]
            first_segment = first_segment + "(/"
        if intermediate_segment.strip()[-2:] == "/)":
            intermediate_segment = intermediate_segment.strip()[:-2]
            last_segment = "/)" + last_segment

        string = "%s %s%%val %s" % (first_segment, intermediate_segment, last_segment)
    return string


# Routine to know if we are entering or exiting a routine
# If the line corresponds to the start of a routine, returns the name of this routine
# If the line corresponds to the end of a routine, returns None
# If is neither one of these cases, returns the same routine
def in_which_block(_line, current_block, vault=None):
    _line = _line.strip().lower()
    _line = remove_literals(_line)
    _pattern_sub = r"subroutine *(\w*) *\(*"
    _pattern_func = r"^[^:]*function *(\w+) *\(*"
    _pattern = None
    if _line.count("subroutine"):
        _pattern = _pattern_sub
    elif _line.count("function"):
        _pattern = _pattern_func
    # We are changing block
    if _pattern is not None:
        # We are exiting a block
        if _line.find("end") == 0:
            current_block.routine = None
        # We are entering a new block
        else:
            match = re.search(_pattern, _line, flags=re.IGNORECASE)
            # cases like variables names that contains "function" in it will not match
            if match is not None:
                block_name = match.groups(0)[0].strip()
                if vault:
                    try:
                        procedure = vault.get_procedure(block_name, module_name=current_block.module_name)
                        current_block.routine = procedure
                    except ProcedureNotFound:
                        current_block.routine = None
                else:
                    current_block.routine = block_name
    return current_block


def in_which_type_definition(_line, type_def):
    _line = _line.strip().lower()
    _line = remove_literals(_line)
    _pattern_type = "^ *type[^(].*::(.*)"
    _pattern_type_alt = "^ *type[^(](.*)"
    _pattern_type_end = "^ *end *type"

    # We are changing block
    if type_def is None:
        m = re.match(_pattern_type, _line, re.I)
        if m:
            type_name = m.group(1).strip()
            return type_name
        n = re.match(_pattern_type_alt, _line, re.I)
        if n:
            type_name = n.group(1).strip()
            return type_name
    else:
        m = re.search(_pattern_type_end, _line, re.I)
        if m:
            return None
    return type_def


def literal_mask(string):
    splitted_string = [c for c in string]
    length = len(splitted_string)
    mask = []
    inside_literal = False
    marker = None
    for i in range(length):
        if splitted_string[i] == "\"":
            if not inside_literal:
                inside_literal = True
                marker = "\""
                mask.append(False)
            else:
                if marker == "\"":
                    inside_literal = False
                    marker = None
                    mask.append(False)
                elif marker == "'":
                    mask.append(True)
                else:
                    raise AssertionError("Should not happen!")
        elif splitted_string[i] == "'":
            if not inside_literal:
                inside_literal = True
                marker = "'"
                mask.append(False)
            else:
                if marker == "'":
                    inside_literal = False
                    marker = None
                    mask.append(False)
                elif marker == "\"":
                    mask.append(True)
                else:
                    raise AssertionError("Should not happen!")
        else:
            if not inside_literal:
                mask.append(False)
            else:
                mask.append(True)

    return mask


def remove_comments(string):
    # Trying to use a rarely used character ({) for the temporal replacement
    if string.count("!"):
        if string.count('$OMP'):
            return string
        # In the case where there's quotes in the text, we should ignore the exclamations inside
        if not string.count("'") and not string.count('"'):
            string = string[:string.find("!")]
        else:
            mask = literal_mask(string)
            # Looking for
            splitted_string = [c for c in string]
            splitted_string = ["{" if c == "!" and mask[i] else c for i, c in enumerate(splitted_string)]
            string = "".join(splitted_string)

            if string.find("!") >= 0:
                string = string[:string.find("!")]
            string = string.replace("{", "!")

    return string


def find_argument_intervals(string, _field=None):
    # Given a line with a call to a function, give the index intervals of the different arguments.

    # Mask to avoid scanning inside strings
    mask = literal_mask(string)

    # Find all opening parenthesis
    op = [i for i, x in enumerate(string) if x == "(" and not mask[i]]
    # Find all closing parenthesis
    cp = [i for i, x in enumerate(string) if x == ")" and not mask[i]]

    if not op and not cp:
        return 0, None

    # Keep only arguments part
    arguments_text = string[op[0] + 1:cp[-1]]
    # Find remaining parenthesis
    start = op[0] + 1
    end = cp[-1]
    op = [i for i, x in enumerate(arguments_text) if x == "("]
    cp = [i for i, x in enumerate(arguments_text) if x == ")"]
    osp = [i for i, x in enumerate(arguments_text) if x == "[" and not mask[i]]
    csp = [i for i, x in enumerate(arguments_text) if x == "]" and not mask[i]]

    # Find commas

    splitted_line = [c for c in arguments_text]

    commas = []
    mask = literal_mask(arguments_text)
    for i, c in enumerate(splitted_line):
        # Had to add this to avoid problems with commas inside strings"
        if c == ",":
            if not mask[i]:
                commas.append(i)

    # For existing commas:
    argument_separating_commas = []
    for n in commas:
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x < n]) == len([x for x in cp if x < n]):
            if len([x for x in osp if x < n]) == len([x for x in csp if x < n]):
                splitted_line[n] = "#"
                argument_separating_commas.append(n)

    # Displacing the argument_separating_commas to match the original line
    argument_separating_commas = [_x + start + 1 for _x in argument_separating_commas]
    # Adding first and last elements
    argument_separating_commas = [start] + argument_separating_commas + [end + 1]

    # Constructing intervals
    intervals = [(argument_separating_commas[i], argument_separating_commas[i + 1] - 1) for i in
                 range(len(argument_separating_commas) - 1)]
    if _field is not None:
        if len(intervals) <= _field:
            return len(intervals), None
        else:
            return len(intervals), intervals[_field]
    else:
        return len(intervals), intervals


def is_keyword_argument(argument):
    pattern = "^[^()'\"=]*=[^=]"
    if re.search(pattern, argument):
        return True
    else:
        return False


def add_function_to_argument(string, field_index, _function, number_of_fields=None):
    # Given a line with a call, add a call to a function in a specific field by index.
    list_of_key = ["dim", "mask", "kind", "back"]
    num_arguments, interval = find_argument_intervals(string, field_index)
    if num_arguments <= field_index:
        return string
    if number_of_fields:
        if number_of_fields is not num_arguments:
            return string
    first_segment = string[:interval[0]]
    intermediate_segment = string[interval[0]:interval[1]]
    last_segment = string[interval[1]:]

    # The arguments containing a literal are discarded (Probably a bad workaround)
    if intermediate_segment.count("'"):
        #     return _line
        # raise ValueError("Probably the argument should not contain quotes at this point")
        warnings.warn("Probably the argument should not contain quotes at this point:\n%s" % intermediate_segment)

    # The arguments that already contain a call to the function are discarded (Probably a bad workaround)
    if re.match(r"^ *%s *\(.*" % _function, intermediate_segment):
        # raise ValueError("Probably the argument should not contain %s at this point" % _function)
        return string

    # If there are named arguments, add the function to the actual argument
    # (if the equal sign is inside a parenthesis, this should not be done)
    named_argument_workaround = False

    if intermediate_segment.count("="):
        if intermediate_segment.count("("):
            if intermediate_segment.find("=") < intermediate_segment.find("("):
                named_argument_workaround = True
        else:
            named_argument_workaround = True

    if named_argument_workaround:
        parts = intermediate_segment.split("=")
        intermediate_segment = "%s=%s( %s )" % (parts[0], _function, parts[1])
    else:
        intermediate_segment = "%s( %s )" % (_function, intermediate_segment)
    string = "%s%s%s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_rpe_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "rpe", number_of_fields)
    return string


def add_rpe_var_to_arguments(string):
    num_arguments, intervals = find_argument_intervals(string)
    for i in range(num_arguments):
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        string = "%s rpe_var( 52, %s ) %s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_real_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "real", number_of_fields)
    return string


def has_intrinsics(string, return_intrinsic_name=False):
    string = string.lower()

    for key in list_of_key:
        if string.count(key):
            key_match = r",\s*[ ,(+-]%s *= *\w" % key
            string = re.sub(key_match, "", string)
    # We just need to check the assignation, not the variable name
    if string.count("="):
        string = string.split("=")[1].strip()

    # Split parenthesis opening
    if not string.count("("):
        return False
    # piece_to_check = string.split("(")[0].strip()

    for intrinsic in list_of_intrinsics:
        if re.search(r'\s*%s\s*\(' % intrinsic, string):
            if return_intrinsic_name:
                return intrinsic
            return True
    return False


# Check if a string is a number and returns the type
def is_a_number(string):
    if string.count("="):
        string = string.split("=")[1]

    # Check if is just white spaces
    if not string.strip():
        return False
    else:
        string = string.lower().strip()

    # Remove unary operator if present
    if string[0] == '-' or string[0] == '+':
        string = string[1:]

    name_of_intrinsic = has_intrinsics(string, True)
    if name_of_intrinsic:
        string = string.replace(name_of_intrinsic, "")
        string = string.strip()
        string = string[1:-1]
        string = string.strip()

    # Checks for number in exponential notation
    exp = re.match("^" + pattern_exp_DPE_notation + "$", string)
    # Checks for number with precision specified
    real_with_prec = re.match(pattern_real_with_prec, string)
    if exp or real_with_prec:
        # In case the number is a hardcoded real with a type specification,
        # check the lookup table to know to which type it corresponds and return it
        try:
            new_type = lookup_table[string.split("_")[1]]
        except IndexError:
            return working
        return new_type

    # Check for pure integer and real numbers
    try:
        float(string)
        if string.isdigit():
            return "integer"
        else:
            return working
    except ValueError:
        return False


def find_kind_of_parameter_exception(arguments):
    if not arguments.count(","):
        argument = arguments.split("=")[1]
        num_type = is_a_number(arguments)
        if num_type:
            return "is_a_number"
        if has_operations(argument):
            type_of_exception = 'has_operation'
        elif has_intrinsics(argument):
            type_of_exception = 'has_intrinsics'
        else:
            type_of_exception = "regular"
    else:
        if arguments.count("=") > 1:
            type_of_exception = "multiple_variables"
        elif arguments.count("(/"):
            type_of_exception = "array_initialization"
        elif arguments.count("RESHAPE(["):
            type_of_exception = "array_reshape"
        elif re.search(r"= *\[.*\]", arguments):
            type_of_exception = "array_reshape"
        elif arguments.count("rpe_var("):
            type_of_exception = "already_fixed"
        else:
            raise TypeError("Unexpected kind of exception")
    return type_of_exception


def remove_brackets_content(string):
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    _new_line = []
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x <= n]) == len([x for x in cp if x < n]):
            _new_line += [string[n]]
    return "".join(_new_line)


def separate_line_container_and_arguments(string):
    # Given a line containing a call to a write function, it returns the list of the arguments and the
    # remaining of the line without the arguments.

    # Locate start of write call
    loc = re.search(r"(\W|^)(write)\W", string, re.I).start(2)
    # Save the part of the string that goes before the call if any
    pre_text = string[:loc]
    # Keep the rest of the string
    post_text = string[loc:]
    # Temporarily replace write statement with all the parameters
    write_statement = close_brackets(post_text)
    line = post_text.replace(write_statement, "write()")

    # Find match (?)
    pattern = r"^[^\!]* *write\([^()]*\) (.*)"
    matches = re.search(pattern, line, re.IGNORECASE)
    to_replace = matches.groups(0)[0]

    # Prepare line container
    if to_replace.strip():
        line_container = line.replace(to_replace, "###TOREPLACE###")
    else:
        line_container = line
        return line_container, []

    arguments = to_replace.strip()

    if not arguments:
        return line_container, []

    # Fix array constructors
    if arguments.find("(/") == 0:
        arguments = arguments[2:-2].strip()
        line_container = line_container.replace("###TOREPLACE###", "(/ ###TOREPLACE### /)")

    if all_is_inside_parenthesis(arguments):
        arguments = arguments[1:-1].strip()
        line_container = line_container.replace("###TOREPLACE###", "( ###TOREPLACE### )")

    arguments = split_write_arguments(arguments)

    line_container = pre_text + line_container.replace("write()", write_statement)
    return line_container, arguments

    # Possibly useful in the future
    # implicit_for_pattern = "WRITE(.*, *\w* *=[^,]*, *[\w0-9-+]* *\))"
    # matches = re.search(implicit_for_pattern, string, re.IGNORECASE)
    # if matches:
    #     implicit_for_part = "".join(
    #         list(reversed(close_inverse_brackets("".join(list(reversed(matches.groups()[0])))))))
    #     return []
    #
    # implicit_for_pattern_with_step = "WRITE(.*, *\w* *=[^,]*, *[\w0-9-+]* *, *[\w0-9-+]* *\))"
    # matches = re.search(implicit_for_pattern_with_step, string, re.IGNORECASE)
    # if matches:
    #     implicit_for_part = "".join(
    #         list(reversed(close_inverse_brackets("".join(list(reversed(matches.groups()[0])))))))
    #
    #     return []


def split_write_arguments(string):
    # Given a line containing a call to a read function, it returns the list of the arguments
    # pattern = "^[^\!]* +read\([^()]*\) *(.*)"
    # matches = re.search(pattern, string, re.IGNORECASE)
    arguments = string

    scan_quotation = [arguments[:i].count("'") for i, c in enumerate(arguments)]
    scan_quotation2 = [arguments[:i].count("\"") for i, c in enumerate(arguments)]
    scan_op_brackets = [arguments[:i].count("(") for i, c in enumerate(arguments)]
    scan_cl_brackets = [arguments[:i].count(")") for i, c in enumerate(arguments)]
    scan_brackets = [i - j for i, j in zip(scan_op_brackets, scan_cl_brackets)]
    splitted_text = [c for c in arguments]
    splitted_text = [
        "#" if c == "," and (scan_quotation[i] % 2 == 0 and scan_quotation2[i] % 2 == 0) and not scan_brackets[i] else c
        for i, c in enumerate(splitted_text)]
    arguments = "".join(splitted_text)
    arguments = [a for a in arguments.split("#") if a.strip()]
    return arguments


def split_read_arguments(string):
    arguments = []

    # First check for structure
    match_struct = re.findall(pattern_struct, string, re.I)
    for struct in match_struct:
        arguments.append(struct)
        string = string.replace(struct, "")
        if not string:
            return arguments
    # Now for vectors
    match_vec = re.findall(pattern_vec, string, re.I)
    for vec in match_vec:
        arguments.append(vec)
        string = string.replace(vec, "")
        if not string:
            return arguments

    # now that all vectors are pruned parenthesis can be deleted
    string = string.replace("(", "")
    string = string.replace(")", "")
    # Check for lists
    if string.count("="):
        # remove index of lists, that are integers
        indexes = re.findall(r"\w+=\d+", string)
        for index in indexes:
            string = string.replace(index, "")
    # The rest is just scalar
    for scalar in string.split(","):
        # add just if is not empty
        if not scalar:
            continue
        arguments.append(scalar)
    return arguments


def resulting_type(type_a, type_b):
    types = ["rpe_var", "complex", double, single, working, "integer", "char", "logical"]
    for t in types:
        if type_a == t:
            return t
        if type_b == t:
            return t


def combine_types(list_of_types):
    to_return = list_of_types[0]
    for t in list_of_types:
        to_return = resulting_type(to_return, t)
    return to_return


def calls_to_routines(string):
    pattern = "(^|[ \t]*)call +"
    matches = [(m.start(0), m.end(0)) for m in re.finditer(pattern, string, flags=re.I)]
    if matches:
        pieces = [string[start:].strip() for start, end in matches]
        return [close_brackets(piece) for piece in pieces if piece.count("(")]
    else:
        return []


def remove_literals(string):
    mask = literal_mask(string)
    modified_string = "".join(
        [c if not mask[i] else " " for i, c in enumerate(string)])
    return modified_string


def is_pointer_assignment(string):
    pattern = ".* => .*"
    mask = literal_mask(string)
    match = re.search(pattern, string, flags=re.I)
    if match and not mask[string.index("=>")]:
        return True
    return False


def get_pointer_target(line, routine, vault, check):
    try:
        # Ignore the if condition. This is not an ideal solution, but the easier for the moment
        line = remove_if_condition(line)
        if line.count("::"):
            line = line.split("::")[1]
        pointers = line.strip().split("=>")
        ptr = retrieve_variable(pointers[0], routine=routine, vault=vault)
        target = retrieve_variable(pointers[1], routine=routine, vault=vault)
        return ptr, target
    except AssertionError:
        # case in which the pointer is assigned to NULL()
        return [None, None]
        # return [Ptr, None]


def calls_to_functions(string, vault=None, current_routine=None):
    from .Classes import Function, Interface
    functions = calls_to_procedure(string, procedure_type=Function, vault=vault)
    interfaces = calls_to_procedure(string, procedure_type=Interface, vault=vault)
    # Assert that what was selected where functions
    to_return = functions + interfaces
    to_return = [fun for fun in to_return if is_call_to_function(fun, routine=current_routine, vault=vault)]
    return to_return


def calls_to_procedure(string, procedure_type, vault=None):
    from .Classes import Function, Interface
    # line without literals to avoid parenthesis inside strings and other ugly stuff
    no_literal_line = remove_literals(string)
    to_return = []
    if re.search(".*function .*", string, re.IGNORECASE):
        return to_return
    if re.search(".*subroutine.*", string, re.IGNORECASE):
        return to_return
    pattern = r"(\w+) *\("
    matches = [m for m in re.finditer(pattern, no_literal_line, flags=re.I)]
    if matches:
        if vault:
            for m in matches:
                fun_name = m.groups(0)[0]
                try:
                    fun = vault.get_procedure(fun_name, procedure_type=procedure_type)
                    # In case the function name corresponds to a interface, look for the corresponding function.
                    # if isinstance(fun, Interface):
                    #     function_call = close_brackets(string[m.start(1):])
                    #     arguments = find_call_arguments(function_call)
                    #     try:
                    #         fun = match_arguments_and_interface(arguments, fun, vault=vault)
                    #     # Ignore AssertionError and MemberNotFound exceptions.
                    #     except AssertionError:
                    #         fun = None
                    #     except MemberNotFound:
                    #         fun = None

                except ProcedureNotFound:
                    fun = None
                # TODO: Think a way to avoid having nf90 exceptions inside routines
                if fun or (procedure_type == Function and fun_name.lower().find("nf90_") == 0):
                    function_call = close_brackets(string[m.start(1):])
                    to_return.append(function_call)
        else:
            for m in matches:
                function_call = close_brackets(string[m.start(1):])
                to_return.append(function_call)

    return to_return


def find_called_routine(string):
    pattern = r"^[^\!']*[\t ]*call +([^\(]*)"
    match = re.search(pattern, string, flags=re.I)
    if match:
        return match.groups()[0].strip().lower()
    else:
        return None


def find_called_function(string):
    return string[:string.find("(")].strip()


def find_call_arguments(string):
    n_arg, ints = find_argument_intervals(string)
    if n_arg == 0:
        return None
    if ints:
        arguments = [string[interval[0]:interval[1]].strip() for interval in ints]
        return [x for x in arguments if x]
    else:
        return None


def all_is_inside_parenthesis(string):
    if string.strip()[0] == "(":
        try:
            if "c%s" % string.strip() == close_brackets("c%s" % string.strip()):
                return True
            else:
                return False
        except ValueError:
            return False
    else:
        return False


# Try and remove intrinsics from around a variable
def clean_from_intrinsics(contents, vault=None):
    if vault:
        intrinsics = vault.intrinsics
    else:
        intrinsics = list_of_intrinsics
    pattern = r"(\w+) *\("
    m = [_m for _m in re.finditer(pattern, contents)]
    ret_val = contents
    if m:
        for mat in reversed(m):
            function = contents[mat.start(1):mat.end(1)]
            if function in intrinsics:
                if function == "null":
                    continue
                arg_of_func = contents[mat.end(0):]

                try:
                    arg_of_intrinsic = close_brackets(arg_of_func)
                    # Bad workaround to fix the case in which the argument is not a vector
                    if arg_of_intrinsic.count(",("):
                        arg_of_intrinsic = arg_of_intrinsic.split(",")[0]
                    intrinsic_call = close_brackets(contents[mat.start(1):])
                except ValueError:
                    return contents
                ret_val = contents.replace(intrinsic_call, arg_of_intrinsic)
                if has_intrinsics(ret_val):
                    try:
                        ret_val = clean_from_intrinsics(ret_val)
                    except RecursionError:
                        return contents
                else:
                    return ret_val
            elif function == "real":
                arguments = find_call_arguments(contents)
                return arguments[0]
    return ret_val


def clean_contents(contents, vault=None):
    if vault:
        intrinsics = vault.intrinsics
    else:
        intrinsics = ["reshape", "min", "max", "sum", "log", "maxval", "minval", "abs", "size", "nint"]

    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    if re.match(r"\(/(.*)/\)", contents):
        contents = re.match(r"\(/(.*)/\)", contents).groups(1)[0]

    if contents[0] == "(":
        if "dummy%s" % contents == close_brackets("dummy%s" % contents):
            contents = contents[1:-1]

    if re.match(r"^([^\()]*) *\)+ *$", contents):
        contents = re.match(r"^([^\()]*) *\)+ *$", contents).groups(1)[0]

    if re.match(r".*\) *\(.*\)", contents):
        match = re.match(r"(.*\)) *\(.*\)", contents)
        contents = match.groups(0)[0]

    pattern = r"(\w+) *\("
    m = [_m for _m in re.finditer(pattern, contents)]
    if m:
        for mat in reversed(m):
            prefix = contents[:mat.start(1)]
            part_to_modify = contents[mat.start(1):]
            to_replace = close_brackets(part_to_modify)
            replacement = to_replace[:to_replace.find("(")]
            if replacement in intrinsics and is_call_to_function(contents):
                continue
            contents = "%s%s" % (prefix, part_to_modify.replace(to_replace, replacement))
    if contents.count("(/"):
        # Is temporal array
        contents = contents.replace("(/", "")
        contents = contents.replace("/)", "")
    return contents


def remove_if_condition(_line):
    if _line.replace(" ", "").lower().strip().find("if(") == 0:
        condition = close_brackets(_line)
        _line = _line.replace(condition, "")
    return _line


def is_call_to_function(contents, routine=None, vault=None):
    """
    This function first does a first filter checking that the syntax is similar to a function call and afterwards
    uses one of the two possible approaches, depending on having a vault available or not.
    """
    pattern = r"[-+]* *(\w+) *\("
    m = re.match(pattern, contents)
    if m:
        if contents == close_brackets(contents):
            arguments = find_call_arguments(contents)
            # In case there are no arguments, for sure it is a function and not an array
            if len(arguments) == 0:
                return True
            # Otherwise, use one of the two approaches
            if vault is not None:
                return is_call_to_function_vault(contents, routine=routine, vault=vault)
            else:
                return is_call_to_function_no_vault(contents)
    return False


def is_call_to_function_no_vault(contents):
    """
    This function tries to identify if a certain string is a call to a function or it's an array.
    We don't rely on nothing else than the string itself.
    """
    arguments = find_call_arguments(contents)
    for argument in arguments:
        if argument == ":":
            return False
    # TODO: This will wrongly identify arrays with integer incices as functions.
    return True


def is_call_to_function_vault(contents, routine, vault):
    """
    This function tries to identify if a certain string is a call to a function or it's an array.
    We rely on knowing in which routine we are and having a vault available.
    """
    pattern = r"[-+]* *(\w+) *\("
    function_name = re.match(pattern, contents).group(1)

    # Check if the name matches with the name of a variable in the actual routine
    # TODO: ideally routine should never be None. In some cases it happens
    if routine is not None:
        if any([v for v in routine.variables if v.name == function_name]):
            return False

    # Search for a procedure with the same name
    try:
        fun = vault.get_procedure(function_name)
    except ProcedureNotFound:
        fun = None
    if fun:
        return True

    # Search for a variable with the same name
    try:
        var = retrieve_variable(function_name, routine, vault)
    except AssertionError:
        var = None
    if var:
        if var.name == var.routine.name:
            print("Wait a minute!")
        return False
    # TODO: Here we are assuming that if it has not been identified neither as a function existing in the vault
    # TODO: nor a variable existing in the vault it corresponds to an external function. In case this assumption is not
    # TODO: True, this should be fixed.
    return True


def type_of_real_cast(string):
    # Leave only the arguments
    arguments = find_call_arguments(string)
    if len(arguments) == 1:
        return working
    elif len(arguments) == 2:
        argument, type = arguments
        return lookup_table[type]


def type_of_contents(contents, routine=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions

    if contents == "*":
        return None
    if not vault:
        vault = load_vault()
    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    match = re.search(pattern_array, contents)
    if match:
        contents = contents.replace("(/", "")
        contents = contents.replace("/)", "")

    if all_is_inside_parenthesis(contents):
        contents = contents.strip()[1:-1]
        # raise AssertionError("Remove the first and last bracket")

    # Case with call to a function with array arguments after
    match = re.search(r"(\w* *\([^=]*\)) *(\(.*\))", contents)
    if match:
        contents = match.group(1)

    contents = contents.lower().strip()
    if contents[0] in ["'", '"'] or contents[:4] == 'trim':
        return "char"

    is_a_comparison = split_elements_in_comparison(contents)
    if is_a_comparison is not None:
        return "logical"
        # ret_type = type_of_contents(is_a_comparison[0], routine=routine, vault=vault)
        # for var in is_a_comparison[1:]:
        #     if ret_type == type_of_contents(var, routine=routine, vault=vault):
        #         continue
        #     else:
        #         warnings.warn("Not able to find type! :(  |%s|" % contents)
        #         break
        # return ret_type

    contents = contents.replace(" ", "").strip()

    if is_call_to_function(contents):
        fun_name = find_called_function(contents)
        if fun_name in vault.intrinsics:
            if fun_name in ["reshape", "min", "max", "sum", "log", "log10", "maxval", "minval", "abs", "sqrt", "maxloc",
                            "minloc", "huge", "transfer", "sign"]:
                try:
                    return type_of_contents(find_call_arguments(contents)[0], routine=routine, vault=vault)
                except RecursionError as rec:
                    print('Recursion error: {}'.format(rec.args[0]))
            elif fun_name in ["int", "nint", "count", "size", "mod", "umask", "len", "len_trim", "count"]:
                return "integer"
            elif fun_name in ["real", "float", "sin", "cos", "acos", "exp", "cosh"]:
                if fun_name == "real":
                    return type_of_real_cast(string=contents)

                return working
            elif fun_name in ["cmplx"]:
                return "complex"
            elif fun_name in ["trim", "ctime", "adjustl"]:
                return "char"
            elif fun_name in ["present"]:
                return "logical"
            elif fun_name in ["null"]:
                return "pointer"
            elif fun_name in ["rpe"]:
                return "rpe_var"
            else:
                raise ValueError("This intrinsic should be considered too: %s" % fun_name)
        else:
            from .Classes import Interface
            try:
                fun = vault.get_procedure(fun_name)
                if isinstance(fun, Interface):
                    return fun.procedures[0].type
                else:
                    return fun.type
            except ProcedureNotFound:
                pass
            # warnings.warn("If its a function should be on the library or an intrinsic, isn't it? %s|nLooking if its a variable... " % contents)
            # return type_of_contents(fun_name, routine=routine, vault=vault)

    if has_operations(contents):
        split_elements = split_elements_in_operations(contents)
        try:
            types = [type_of_contents(element, routine, vault) for element in split_elements]
            return combine_types(types)
        except RecursionError as rec:
            print('Recursion error: {}'.format(rec.args[0]))
    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        try:
            return type_of_contents(contents, routine=routine, vault=vault)
        except RecursionError as rec:
            print('Recursion error: {}'.format(rec.args[0]))

    contents = clean_contents(contents, vault)

    if re.match("^" + pattern_exp_DPE_notation + "$", contents):
        return working

    if contents.find("nf90_") == 0:
        return "integer"
    if contents.find("set_nf90_") == 0:
        return "integer"

    if contents.count("'") or contents.count("\""):
        return "char"
    if contents.count(","):
        # If there are commas at this point, it must correspond with a temporal array
        # Array elements should have the same type, so we will assume that all the elements
        # have the same type than the first element
        first_element = split_attributes(contents)[0]
        return type_of_contents(first_element, vault=vault)
    if contents.count(","):
        raise ValueError("Argument should not contain commas at this point %s" % contents)

    if contents.count("(") or contents.count(")"):
        warnings.warn("WARNING: Argument should not contain brackets at this point % s" % contents)
        contents = contents.replace("(", "")
        contents = contents.replace(")", "")

    if contents.count(".false.") or contents.count(".true."):
        return "logical"

    # Checks if it is a number
    num_type = is_a_number(contents)
    if num_type:
        return num_type

    if vault:
        abs_content = contents
        # in case a variable's name is preceded from unary operator
        if contents[0] == '-' or contents[0] == '+':
            abs_content = contents[1:]
        if re.search(r"^\w+$", abs_content):
            coincidences = [v for v in vault.get_variable(abs_content)]
            if len(coincidences) == 1:
                return coincidences[0].type
            elif len(coincidences) > 1:
                ctypes = [v.type for v in coincidences]
                if elements_are_equal(ctypes):
                    return ctypes[0]

        if routine:
            from .Classes import Routine, ModuleMain, Function
            if isinstance(routine, Routine) or isinstance(routine, ModuleMain) or isinstance(routine, Function):
                procedure = routine
            else:
                procedure = vault.get_procedure(routine)

            for variable in procedure.variables:
                if contents == variable.name.lower():
                    return variable.type
            for variable in procedure.accessible_variables():
                if contents == variable.name.lower():
                    return variable.type

            # Try if there's a function with this name
            try:
                fun = vault.get_function(contents)
                return fun.type
            except ProcedureNotFound:
                pass

    if contents.count("%"):
        # TODO: Strange way of handling errors. Check if there's a better way of doing it
        if contents.split("%")[-1] == "val":
            return double

        member = get_structure_member(contents, routine, vault=vault)
        return member.type
        # return type_of_derived_type_element(contents, routine=routine, vault=vault)

    if contents.count(">") or contents.count("<") or contents.count("=="):
        return "logical"

    if vault:
        if contents.count("mpi_"):
            return "integer"
        elif contents == "rpe":
            return "rpe_var"
        else:
            warnings.warn("Not able to find type! :(  |%s|" % contents)
    else:
        pass

    warnings.warn("Did not find type of: %s" % contents)
    # raise AssertionError("Did not find type of: %s" % contents)


def dimension_of_contents(contents, routine=None, vault=None):
    # TODO: Improve this function to work properly in all cases.
    contents = remove_literals(contents)
    contents = contents.lower()
    # Adding a workaround for casts
    if is_call_to_function(contents):
        fun = find_called_function(contents)
        if fun == "real":
            contents = find_call_arguments(contents)[0]
        elif fun == "sum":
            if contents.count("dim="):
                dim = dimension_of_contents(find_call_arguments(contents)[0], routine=routine, vault=vault)
                if dim:
                    return dim - 1
                else:
                    return dim
            else:
                contents = find_call_arguments(contents)[0]
        elif fun == "sqrt":
            contents = find_call_arguments(contents)[0]
    if has_operations(contents):
        elements = split_elements_in_operations(contents)
        dimensions = [dimension_of_contents(e, vault=vault) for e in elements]
        dimensions = [d for d in dimensions if d is not None]
        if dimensions == []:
            dimensions = [0]
        return max(dimensions)
    if is_a_number(contents):
        return 0
    if contents.count(":"):
        return contents.count(":")
        # Case with call to a function with array arguments after
    match = re.search(r"\(/.*/\)", contents)
    if match:
        return 1
    if re.search(r"\w *\([0-9\w]+\)", contents):
        return 0
    try:
        var = retrieve_variable(contents, routine=routine, vault=vault)
        # Checking if we have the var + slice specification... if it is an slice and does not have any colon
        # then it must be an scalar
        if re.search(r"%s *\(.*\)" % var.name, contents):
            return 0
        return var.dimension
    except AssertionError:
        return None


# Retrieve variable corresponding to a structure member
def get_structure_member(contents, routine, vault):
    member = contents.split("%")[-1]
    coincidences = [v for v in vault.get_variable(member) if v.is_member_of is not None]
    if len(coincidences) == 1:
        # Just one structure has such a member
        return coincidences[0]
    else:
        # More than one structure has such a member
        for match in coincidences:
            index = 2
            member_of = match.is_member_of
            # Checks the different member going backward, and looks at the compatibility with available structures
            while index <= len(contents.split("%")):
                members = contents.split("%")[-index]
                # Structures which has member of the same type
                struct = [v for v in vault.get_variable(members) if v.type == member_of]
                # No structures correspond to this match, go to the next
                if len(struct) == 0:
                    break
                if len(struct) == 1:
                    # Just one structure matches, but we need to check that the parent structure type coincides
                    if struct[0].is_member_of is not None:
                        member_of = struct[0].is_member_of
                        index = index + 1
                    # One match is found and there is no parent structure
                    else:
                        return match
                # More than one struct is found, check parent
                if len(struct) > 1:
                    # Reached the root member we still have more than a match: check if the routine coincide
                    if index == len(contents.split("%")):
                        struct_filter = [s for s in struct if s.routine == routine]
                        if len(struct_filter) == 1:
                            return match
                    index = index + 1
        # Something went wrong in the search
        raise MemberNotFound


def retrieve_variable(contents, routine=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions
    if not vault:
        vault = load_vault()

    contents = contents.replace(" ", "").lower().strip()

    try:
        purged_contents = clean_from_intrinsics(contents)
        contents = purged_contents
    except ValueError:
        print("Not able to clean content ", contents, " from intrinsics")

    contents = clean_contents(contents, vault)

    if re.search(r"^\w+$", contents):
        coincidences = [v for v in vault.get_variable(contents)]
        if len(coincidences) == 1:
            return coincidences[0]
    if contents.count("%val"):
        contents = contents.replace("%val", "")
    if contents.count("%"):
        try:
            return get_structure_member(contents, routine, vault=vault)
        except MemberNotFound:
            raise AssertionError("Could not retrieve variable %s of structure", contents)

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return retrieve_variable(contents, routine=routine, vault=vault)
    if routine:
        from .Classes import Routine, ModuleMain, Function
        if isinstance(routine, Routine) or isinstance(routine, ModuleMain) or isinstance(routine, Function):
            procedure = routine
        else:
            procedure = vault.get_procedure(routine)

        for variable in procedure.variables:
            if contents == variable.name.lower():
                return variable
        for variable in procedure.accessible_variables():
            if contents == variable.name.lower():
                return variable

    raise AssertionError("Did not find this variable: %s" % contents)


def type_of_derived_type_element(contents, routine=None, derived_type=None, vault=None):
    if not vault:
        vault = load_vault()

    struct = contents[:contents.find("%")]
    var = contents[contents.find("%") + 1:]
    if var == "val":
        return working

    if not derived_type:
        derived_type = type_of_contents(struct, routine=routine, vault=vault)

    if derived_type is None:
        raise AssertionError("Did not type of %s" % contents)

    if var.count("%"):

        struct = var[:var.find("%")]
        derived_type = type_of_derived_type_element(struct, routine=routine, derived_type=derived_type, vault=vault)
        var = var[var.find("%") + 1:]
        if var == "val":
            return working
        dt = vault.get_derived_type(derived_type)
        try:
            v = dt.get_variable(var)
        except AttributeError:
            pass

        try:
            return v.type
        except AttributeError:
            pass
    else:
        if vault:
            dt = vault.get_derived_type(derived_type)
            try:
                return dt.get_variable(var).type
            except AttributeError:
                pass

        return "unknown"


def load_source_files(path):
    from .Classes import ReadSourceFile
    filelist = [ReadSourceFile(join(path, filename)) for filename in listdir_ord(path) if filename.count(".f90")]
    filedict = {}
    for f in filelist:
        filedict[f.filename] = f
    return filedict


def listdir_ord(path):
    from os import listdir
    file_list = listdir(path)
    file_list.sort()
    return file_list


def load_vault(filename="vault.pkl"):
    try:
        return pickle.load(open(filename, "rb"))
    except IOError:
        return None


def load_intrinsics():
    package_directory = dirname(abspath(__file__))
    # Hardcoded path to file.
    with open(package_directory + "/../AdditionalFiles/fortran_intrinsics") as f:
        return [re.search(r"(\w+)", l).groups(0)[0].lower() for l in f]


def load_list_of_files_to_keep_unmodified(filename=None):
    if filename is None:
        package_directory = dirname(abspath(__file__))
        filename = package_directory + "/../AdditionalFiles/list_of_files_to_keep_unmodified.txt"
    # Hardcoded path to file.
    if isfile(filename):
        print("Loading list of exceptions from %s\n\n" % filename)
        with open(filename) as f:
            return [l.strip() for l in f if l.strip() and l.strip()[0] != "#"]
    else:
        return None


def close_brackets(string):
    started = False
    bracket_counter = 0
    mask = literal_mask(string)

    for _i, _c in enumerate(string):
        if _c == "(" and not mask[_i]:
            bracket_counter += 1
            if not started:
                started = True
        elif _c == ")" and not mask[_i]:
            bracket_counter -= 1
            if not started:
                raise ValueError("This should not happen, check arguments %s" % string)
        if bracket_counter < 0:
            raise ValueError("This should not happen, check arguments %s" % string)
        if started and bracket_counter == 0:
            return string[:_i + 1]

    if string.count("(") + string.count(")") == 0:
        return string
    raise ValueError("This should not happen, check arguments %s" % string)


def elements_are_equal(iterator):
    iterator = iter(iterator)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first == rest for rest in iterator)


def split_attributes(attributes):
    # Description: Given a string with declaration attributes separates the different attributes and
    # gives back a list of attributes
    # Input:
    #       -attributes: string with attributes
    # Output:
    #       -attributes_list: list containg strings with the different attributes
    attributes_list = replace_comma_outside_parenthesis_by_hashtag(attributes)
    attributes_list = [l.strip().replace("#", ",") for l in attributes_list.split(",")]
    return attributes_list


def which_type(attributes):
    # Description: Given a list of attributes, this function returns the type of the variables
    # Input: List of attributes
    # Output: String with type
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    assert isinstance(attributes, list)
    possible_types = ["real", "integer", "type", "char", "logical", "complex", "double precision", "useintrinsic",
                      "public"]
    t0 = attributes[0]
    var_type = None
    for ty in possible_types:
        if t0.lower().find(ty) == 0:
            var_type = ty
    if not var_type:
        return None
    if var_type == "type":
        # It may be the case in which it is a type declaration instead of a variable declaration of a type
        # This workaround is to discard this cases.
        if not t0.count("("):
            return None
        var_type = t0[t0.find("(") + 1:-1].strip()
    if var_type == "public":
        return None
    if var_type == "real":
        pattern = r"real *\(([^\)]*)\)"
        if re.search(pattern, t0, re.I):
            type_specifier = re.search(pattern, t0, re.I).group(1).strip()
            try:
                return lookup_table[type_specifier]
            except KeyError:
                print("Key %s not found in lookup table.", type_specifier)
        else:
            AssertionError("Type specifier not found in real variable declaration")
    return var_type.lower()


def which_dimension(attributes, variable):
    # Description: Given a list of attributes, this function returns the dimension of the variables
    # Input: List of attributes
    # Output: Integer with the number of dimensions
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    elif isinstance(attributes, list):
        pass
    else:
        raise AssertionError("Not a string nor a list, what it is?")
    dim = 0
    for att in attributes:
        if re.search(r"dimension\s*\(\*\)", att, re.I):
            dim = None
            break
        if att.lower().count("dimension"):
            dim = 1 + att.count(",")
            break
    if dim == 0:
        dim = variable.count(":")
        # TODO: Probably not the most robust solution to find the dimension
    if variable.count("="):
        variable = variable.split("=")[0].strip()
    if dim == 0 and variable.count("("):
        dim = variable.count(",") + 1

    return dim


def which_intent(attributes):
    # Description: Given a list of attributes, this function returns the intent of the variables
    # Input: List of attributes
    # Output: String with the intent of the variable, if intent is not specified the default is inout
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    elif isinstance(attributes, list):
        pass
    intent = "inout"
    for att in attributes:
        if att.lower().count("intent"):
            intent = att
            break
    if intent.count("("):
        intent = intent[intent.find("(") + 1:-1].strip().lower()
    return intent


def is_optional(attributes):
    # Description: Given a list of attributes, this function returns a boolean describing if the variable is optional
    # Input: List of attributes
    # Output: Boolean about optionality of variable
    attributes = split_attributes(attributes)
    optional = False
    for att in attributes:
        if att.lower().count("optional"):
            optional = True
            break
    return optional


def is_parameter(attributes):
    # Description: Given a list of attributes, this function returns a boolean describing if the variable is optional
    # Input: List of attributes
    # Output: Boolean about optionality of variable
    attributes = split_attributes(attributes)
    for att in attributes:
        if att.lower().count("parameter"):
            return True
    return False


def is_allocatable(attributes):
    # Description: Given a list of attributes, this function returns a boolean describing if the variable is allocatable
    # Input: List of attributes
    # Output: Boolean about if variable is allocatable
    attributes = split_attributes(attributes)
    for att in attributes:
        if att.lower().count("allocatable"):
            return True
    return False


def is_pointer(attributes, variables):
    # Description: Given a list of attributes, check if pointer is among them
    # Input: List of attributes
    # Output: boolean
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    assert isinstance(attributes, list)
    if "pointer" in attributes or "POINTER" in attributes:
        return True
    return False


def parse_attributes(attributes, variables):
    if isinstance(attributes, str):
        attributes = split_attributes(attributes)
    assert isinstance(attributes, list)
    var_type = which_type(attributes)
    dimension = which_dimension(attributes, variables)
    intent = which_intent(attributes)
    optional = is_optional(attributes)
    allocatable = is_allocatable(attributes)
    parameter = is_parameter(attributes)
    pointer = is_pointer(attributes, variables)
    return var_type, dimension, intent, optional, allocatable, pointer, parameter

def split_declaration_arguments(string):
    # Description: given a string containing arguments, return a list with arguments
    # Input: String with arguments
    # Output: List with arguments
    # TODO: If the string of arguments contains parenthesis, we should keep information about the dimensions.
    # TODO: If the argument has a value declaration, we should get rid of this.

    string = remove_literals(string)
    string = replace_comma_outside_parenthesis_by_hashtag(string)
    string = [l.strip().replace("#", ",") for l in string.split(",")]
    string = [el[:el.find("(")] if el.count("(") else el for el in string]
    string = [el[:el.find("=")].strip() if el.count("=") else el for el in string]
    string = [item.replace(" ", "") for item in string]
    return string


def replace_comma_outside_parenthesis_by_hashtag(string):
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    line = [c for c in string]
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if string[n] == ",":
            if len([x for x in op if x <= n]) != len([x for x in cp if x < n]):
                line[n] = "#"
    string = "".join(line)
    return string


def split_elements_in_operations(string):
    operations = ["+", "-", "*", "/"]
    exponentials_DPE = re.findall(pattern_exp_DPE_notation, string, re.I)
    pattern_pow = r'\*\*\d+\.?'
    # pows = re.findall(pattern_pow, string, re.I)
    # for pow in pows:
    #     string = string.replace(pow, "")
    string.replace("(/", "BEGIN_VECTOR")
    string.replace("/)", "END_VECTOR")
    # Substitute exponential with string
    for index, exponential_DPE in enumerate(exponentials_DPE):
        string = string.replace(exponential_DPE, "EXPONENTIAL_DPE" + str(index))

    string = new_purge_parenthesis(string)

    split_elements = split_elements_by_symbols(string, operations)

    for element in split_elements:
        element.replace("BEGIN_VECTOR", "(/")
        element.replace("END_VECTOR", "/)")

    # Substitute back expontenials.
    for index, exponential_DPE in enumerate(exponentials_DPE):
        sub = "EXPONENTIAL_DPE" + str(index)
        for element in split_elements:
            if element.count(sub):
                split_elements[split_elements.index(element)] = element.replace(sub, exponential_DPE)
    return split_elements


def split_elements_in_comparison(string):
    comparison_operators = [">=", "<=", "==", "/=", ">", "<", ".or.", ".and."]
    elements_in_comparison = []
    string = string.lower()

    string = new_purge_parenthesis(string)

    for operators in comparison_operators:
        if string.count(operators):
            for el in string.split(operators):
                elements_in_comparison.append(el.strip())
            break
    if elements_in_comparison:
        return elements_in_comparison
    else:
        return None


def purge_parenthsis(string):
    # Substitute structures with string
    structures = re.findall(pattern_struct, string)
    for index, struct in enumerate(structures):
        string = string.replace(struct, "STRUCT" + str(index))

    index = 0
    arrays = True
    arr_list = []

    # Checks for all array pattern, and substitute with string
    while arrays:
        arrays = re.findall(pattern_vec, string)
        for array in arrays:
            string = string.replace(array, "ARRAY" + str(index), 1)
            arr_list.append(array)
            index += 1

    # Now is safe to discard brackets
    string = string.replace("(", " ")
    string = string.replace(")", " ")

    # Substitute back the original arrays (reversed in case of arrays of array)
    for index, array in reversed(list(enumerate(arr_list))):
        string = string.replace("ARRAY" + str(index), array)

    # Substitute back the original structures
    for index, struct in enumerate(structures):
        string = string.replace("STRUCT" + str(index), struct)
    return string


def purge_arr_struc(string, ret_rep_list=False):
    replacement_index = 0
    arrays = True
    replacement_list = []
    # Checks for all array pattern, and substitute with string
    while arrays:
        arrays = re.findall(vec, string)
        for array in arrays:
            string = string.replace(array, "REPLACEMENT" + str(replacement_index), 1)
            replacement_list.append(array)
            replacement_index += 1

        # Substitute structures with string
        structures = re.findall(r"\w+%\w+", string)
        for struct in structures:
            string = string.replace(struct, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(struct)
            replacement_index += 1
            arrays = True
    if ret_rep_list:
        return string, replacement_list
    return string


def new_purge_parenthesis(string):
    replacement_index = 0
    arrays = True
    replacement_list = []
    # Checks for all array pattern, and substitute with string
    while arrays:
        arrays = re.findall(vec, string)
        for array in arrays:
            string = string.replace(array, "REPLACEMENT" + str(replacement_index), 1)
            replacement_list.append(array)
            replacement_index += 1

        # Substitute structures with string
        structures = re.findall(r"\w+%\w+", string)
        for struct in structures:
            string = string.replace(struct, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(struct)
            replacement_index += 1
            arrays = True
    functions = re.search(r"\w+\s*\(", string)
    if functions:
        matches = re.finditer(r"\w+\s*\(", string, flags=re.I)
        matches_start = [m.start(0) for m in matches]
        for start in matches_start:
            func = close_brackets(string[start:])
            string = string.replace(func, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(func)
            replacement_index += 1

    # Now is safe to discard brackets
    string = string.replace("(", " ")
    string = string.replace(")", " ")

    # Substitute back the original structures
    for index, replacement in reversed(list(enumerate((replacement_list)))):
        string = string.replace("REPLACEMENT" + str(index), replacement)

    return string


def split_elements_by_symbols(string, symbols):
    functions = calls_to_functions(string)
    dictionary_of_replacements = {}
    if functions:
        for index, function in enumerate(functions):
            dictionary_of_replacements["REPLACEMENT%i" % index] = function
            string = string.replace(function, "REPLACEMENT%i" % index)

    listed_text = [c for c in string]
    parenthesis_counter = 0
    for index, c in enumerate(listed_text):
        if c == "(":
            parenthesis_counter += 1
        elif c == ")":
            parenthesis_counter -= 1
        elif c in symbols and parenthesis_counter == 0:
            listed_text[index] = "#"
    string = "".join(listed_text)
    arguments = [x.strip() for x in string.split("#") if x.strip()]
    for index, argument in enumerate(arguments):
        for key, function in dictionary_of_replacements.items():
            if key in argument:
                arguments[index] = arguments[index].replace(key, function)

    return arguments


def has_operations(string):
    operations = ["+", "-", "*", "/"]
    # Exclude key arguments
    # TODO: Deal with "list comprehension" i.e. : ' (/ (zshift + REAL(ji,wp), ji = 1, jpi_crs*jpj_crs) /)'
    # TODO: This approach does not work
    keyword_pattern = "^ *\w+ *=.*"
    if re.match(keyword_pattern, string, re.I):
        return False
    # Check if string is just white spaces
    if not string.strip():
        return False
    else:
        string = string.lower().replace(" ", "")

    if string.count("'") or string.count('"'):
        return False
    if is_call_to_function(string):
        return False
    # Remove the unary operators
    if string[0] == '-' or string[0] == '+':
        string = string[1:]

    # Replace char that can cause false positive
    string = string.replace("=-", 'EQUAL_MIN')
    string = string.replace("(/", "BEGIN_VEC(")
    string = string.replace("/)", ")")

    # Substitutes arrays and structures, then deletes useless brackets
    string = purge_arr_struc(string)

    # Check for exponential, regex is needed to be sure is not just one of the variables that ends with 'e'
    exp_DPE = re.search(pattern_exp_DPE_notation, string)
    if exp_DPE:
        string = string.replace(exp_DPE.group(), 'EXPONENTIAL_DPE')

    for op in operations:
        if op in string:
            return True
    return False


def match_arguments_and_interface(arguments, interface, routine=None, vault=None):
    argument_types = [type_of_contents(argument, routine=routine, vault=vault) for argument in arguments]
    argument_dimensions = [dimension_of_contents(argument, routine=routine, vault=vault) for argument in arguments]
    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in arguments]
    return match_argument_info_and_interface(argument_types, interface=interface,
                                             argument_dimensions=argument_dimensions,
                                             keyword_arguments=keyword_arguments, vault=vault)


def match_argument_info_and_interface(argument_types, interface, argument_dimensions=None, keyword_arguments=None,
                                      vault=None):
    # This function is intended to take the argument info (types and dimensions) and return the subroutine
    # inside the interface that matches it.
    number_of_arguments = len(argument_types)
    if keyword_arguments is None:
        keyword_arguments = [False for _x in argument_types]
    for procedure in interface.procedures:
        if not vault:
            dummy_arguments = procedure.get_dummy_arguments()
        else:
            dummy_arguments = [var for var in procedure.variables if var.is_dummy_argument and not var.is_member_of]

        try:
            mandatory_arguments = [argument for argument in dummy_arguments if not argument.is_optional]
        except AttributeError:
            print("A problem with dummy argument in procedure %s", procedure.name)
        if len(dummy_arguments) < number_of_arguments:
            continue
        if len(mandatory_arguments) > number_of_arguments:
            continue
        match = True
        for index, argument_type in enumerate(argument_types):
            if match is False:
                continue
            keyword = keyword_arguments[index]
            if keyword:
                var = procedure.get_variable_by_name(keyword)
                if var:
                    dummy_type = var.type
                    dummy_dimension = var.dimension
                else:
                    dummy_type = None
                    dummy_dimension = None
                    match = False
            else:
                dummy_type = procedure.get_variable_by_position(index).type
                dummy_dimension = procedure.get_variable_by_position(index).dimension
            if dummy_type != argument_type:
                if (dummy_type in real_types and argument_type == "rpe_var") or (
                        dummy_type == "rpe_var" and argument_type in real_types):
                    pass
                else:
                    match = False
            if argument_dimensions is not None:
                # At this point we'll just ignore char variables.
                if argument_type == "char":
                    continue
                argument_dimension = argument_dimensions[index]
                if argument_dimension is not None and argument_dimension != dummy_dimension:
                    match = False

        if match:
            return procedure
    raise AssertionError("No matching routine found for interface : %s" % interface.name)


def function_type_from_header(header):
    pattern = "(.*)function"
    real_type_pattern = r"real\((.*)\)"
    match = re.search(pattern, header.strip(), re.IGNORECASE)
    fun_type = remove_brackets_content(match.groups(0)[0].strip().lower())
    if fun_type == "real":
        real_declaration = match.group(1)
        nt = re.match(real_type_pattern, real_declaration, re.I)
        try:
            fun_type = lookup_table[nt.group(1)]
        except AttributeError:
            fun_type = wp

    ret_val = None
    if header.count("result"):
        ret_val = header.lower().split('result')[1]
        ret_val = ret_val.replace('results', '')
        ret_val = ret_val.replace('(', '')
        ret_val = ret_val.replace(')', '')
    if fun_type:
        fun_name = header.lower().split('function')[1]
        fun_name = fun_name.split("(")[0]
        fun_name = fun_name.replace("(", "")
        fun_name = fun_name.strip()
        if not ret_val:
            if match.group(0).count('rpe_var'):
                return 'rpe_var', fun_name
            else:
                return fun_type, fun_name
        else:
            if match.group(0).count('rpe_var'):
                return 'rpe_var', ret_val
            else:
                return fun_type, ret_val
    else:
        if header.count("result"):
            pattern = r"result *\((.*)\)"
            match = re.search(pattern, header, re.IGNORECASE)
            if match:
                return None, remove_brackets_content(match.groups(0)[0].strip().lower())
            else:
                raise ValueError("Not finding function result")

        else:
            pattern = r"function (.*)\("
            match = re.search(pattern, header, re.IGNORECASE)
            if match:
                return None, remove_brackets_content(match.groups(0)[0].strip().lower())
            else:
                raise ValueError("Not finding function result")


def split_vars_by_allocatable(variables):
    allocatables = [var for var in variables if var.type == "rpe_var" and var.allocatable]
    non_allocatables = [var for var in variables if var.type == "rpe_var" and not var.allocatable]
    return allocatables, non_allocatables


def recover_routine_text(variable, file_objects):
    routine = variable.routine
    module = routine.module

    f = file_objects["%s.f90" % module.name]
    to_return = []
    lines = f.lines
    if routine.name == "main":
        for line in lines[:]:
            if line.lower().count("contains"):
                break
            else:
                to_return.append(line)
        to_return.append(lines[-1])
    else:
        printing = False
        for line in lines:
            if not printing:
                if line.lower().count("subroutine"):
                    called_routine = re.search(r"subroutine *(\w*)", line, re.IGNORECASE).group(1)
                    if called_routine.lower() == routine.name.lower():
                        to_return.append(line)
                        printing = True
            else:
                to_return.append(line)
                if line.lower().count("end subroutine"):
                    break

    return to_return


def is_allocated_inside_declaration_scope(var, file_objects):
    text = recover_routine_text(var, file_objects)
    interesting_lines = [line for line in text if
                         re.search(r"allocate *\(.*[^\w]%s[^\w]" % var.name, line, re.IGNORECASE)]
    if interesting_lines:
        return True
    else:
        return False


def split_by_function(variables, fun, files=None):
    pattern = r"[^\w]%s[^\w]"

    vars_true = []
    vars_false = []

    for var in variables:
        appears_in = find_var_elsewhere(var, files)
        found = False
        for f in appears_in:
            appears_in_lines = [line for line in f.lines if
                                re.search(pattern % var.name, line) and not line.count("::")]
            found_array = [True if fun(line) else False for line in appears_in_lines]
            if any(found_array):
                found = True
        if found:
            vars_true.append(var)
        else:
            vars_false.append(var)
    return vars_true, vars_false


def is_namelist_line(line):
    return line.lower().strip().find("namelist") == 0


def split_by_appearance_in_namelist(variables, files=None):
    return split_by_function(variables, is_namelist_line, files=files)


def split_vars_by_main_or_routine(variables):
    in_main = [var for var in variables if var.routine.name == "main"]
    not_in_main = [var for var in variables if var.routine.name != "main"]

    return in_main, not_in_main


# # This function loads a file and applies modifications to the file.
# def postprocess_file(path_to_input_sources, path_to_output_sources, filename, vault):
#     from .Classes import ImplementRPE
#     f = ImplementRPE(join(path_to_input_sources, filename), vault)
#     f.post_process()
#     f.save_modified_file(join(path_to_output_sources, filename))


def find_var_elsewhere(variable, file_objects):
    pattern = r"[^\w]%s[^\w]"

    files = ["%s.f90" % module.name for module in variable.routine.module.used_by]
    files.append("%s.f90" % variable.routine.module.name)
    files_to_search = []
    for f in file_objects:
        if f in files:
            files_to_search.append(file_objects[f])
        if len(files_to_search) == files:
            break

    return [f for f in files_to_search if re.search(pattern % variable.name, f.text)]


def find_file_and_line_of_allocation(variable, file_objects):
    if is_allocated_inside_declaration_scope(variable, file_objects):
        f = file_objects["%s.f90" % variable.routine.module.name]
        for line_index, line in enumerate(f.lines):
            if re.search("allocate.*[^a-zA-Z0-9_%]" + variable.name + "[^a-zA-Z0-9_%]", line, re.IGNORECASE):
                return f, line_index
        raise ValueError("Allocation of variable %s not found" % variable.name)
    else:
        files_where_var_appears = find_var_elsewhere(variable, file_objects)
        for f in files_where_var_appears:
            for line_index, line in enumerate(f.lines):
                if re.search(r"allocate *\(.*[^a-zA-Z0-9_%]" + variable.name + "[^a-zA-Z0-9_%]", line, re.IGNORECASE):
                    return f, line_index
        warnings.warn("Allocation of variable %s not found" % variable.name)
        return None, None


def find_line_in_sourcefile(variable, sourcefile):
    pattern = r"namelist.*[^\w]%s[^\w]" % variable.name
    namelist_name = None
    for line in sourcefile.lines:
        if re.search(pattern, line, re.IGNORECASE):
            pattern2 = r"namelist/(.*)/.*%s[^\w]" % variable.name
            match = re.search(pattern2, line, re.IGNORECASE)
            namelist_name = match.group(1)
            break
    if namelist_name:
        for index, line in enumerate(sourcefile.lines):
            pattern3 = r"read *\(.*[^\w]%s[^\w]" % namelist_name
            if re.search(pattern3, line, re.IGNORECASE):
                return sourcefile, index
    return None, None


def find_file_and_line_of_namelist_reading(variable, file_objects):
    routine = variable.routine
    module = routine.module
    sourcefile = file_objects["%s.f90" % module.name]

    sourcefile, index = find_line_in_sourcefile(variable, sourcefile)

    if sourcefile:
        return sourcefile, index

    other_sources = find_var_elsewhere(variable, file_objects)
    for sourcefile in other_sources:
        sourcefile, index = find_line_in_sourcefile(variable, sourcefile)
        if sourcefile:
            return sourcefile, index

    return None, None


def find_first_use(variable, file_objects):
    routine = variable.routine
    module = routine.module

    sourcefile = file_objects["%s.f90" % module.name]

    # pattern = '(^|\W+)%s(\W+|$)' % variable.name
    pattern = r'\W*%s\W*' % variable.name
    searching = False
    loop_level = 0
    loop_start = None
    in_where_statement = False
    where_statement_start = None

    # Scroll all the lines
    for line_index, line in enumerate(sourcefile.lines):
        line = line.strip().lower()
        # Until it finds the subroutine to which the variables belongs (routine.name)
        if not searching:
            match = re.search(r"(subroutine|function) *(\w*)", line, re.IGNORECASE)
            if match:
                called_routine = match.group(2)
                if called_routine.lower() == routine.name.lower():
                    searching = True
        # In the subroutine
        else:
            # Scroll all the lines searching for the first use
            if re.search(pattern, line, re.IGNORECASE):
                # skip declarations: this is a really bad workaround, but avoids false positive that will block compilation.
                if line.count("::") or line.lower().count("implicit none"):
                    continue
                # Skip the first line
                elif line.count("subroutine"):
                    # Reach the end: the variable has not been used
                    if line.count("end subroutine"):
                        break
                    continue
                # Skip import statements
                elif line.find("use ") == 0:
                    continue
                # Other statement
                elif line.find("data ") == 0:
                    continue
                else:
                    # The variable is found
                    if in_where_statement:
                        # Inside a where statement
                        return sourcefile, where_statement_start
                    else:
                        # In a (nested) loop
                        if loop_level > 0:
                            return sourcefile, loop_start
                        # On a normal line
                        else:
                            return sourcefile, line_index
            elif line.count("end subroutine"):
                break
            # Whenever a 'do loop' starts increment nesting index
            elif re.search(r"^ *do *[\w]* *=", line, re.IGNORECASE):
                if loop_level == 0:
                    loop_start = line_index
                loop_level += 1
            # Whenever a 'do loop' ends decrement nesting index
            # This way match bot the expression 'end do' and 'enddo' but excludes matches like 'enddo loop_name'
            elif re.search(r"^ *end\s*do", line, re.IGNORECASE):
                loop_level -= 1
            # Do the same for where statement
            elif line.find("where") == 0:
                in_where_statement = True
                where_statement_start = line_index
            elif line.find("end where") == 0:
                in_where_statement = False
    return None, None


def insert_variable_precision_specification(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""
    # Add precision specification
    new_line = condition + "    %s%%sbits = emulator_variable_precisions(%i)" % (variable.name, variable.id)
    sourcefile.lines.insert(insertion_line, new_line)

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    sourcefile.lines.insert(insertion_line, new_line)

    sourcefile.rebuild_text()


def insert_variable_usage(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    sourcefile.lines.insert(insertion_line, new_line)

    sourcefile.rebuild_text()


def insert_variable_truncation(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""
    new_line = condition + "    CALL apply_truncation(%s)" % variable.name
    sourcefile.lines.insert(insertion_line, new_line)
    sourcefile.rebuild_text()


def insert_variable_precision_specification_to_allocatable(variable, file_objects, counter_object):
    sourcefile, allocation_line = find_file_and_line_of_allocation(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, allocation_line + 1)
        insert_variable_usage(variable, sourcefile, allocation_line + 1)


def insert_variable_precision_specification_to_namelist_parameter(variable, file_objects, counter_object):
    sourcefile, reading_line = find_file_and_line_of_namelist_reading(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, reading_line + 1)
        insert_variable_usage(variable, sourcefile, reading_line + 1)
    else:
        warnings.warn("Variable %s not found" % variable.name)


def insert_variable_precision_to_routine_variable(variable, file_objects, counter_object):
    sourcefile, first_use_line = find_first_use(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        _line = sourcefile.lines[first_use_line]
        if re.search(r'^\s*if\s*\(', _line, re.I):
            condition = close_brackets(_line)
        else:
            condition = None
        insert_variable_precision_specification(variable, sourcefile, first_use_line, condition=condition)
        insert_variable_truncation(variable, sourcefile, first_use_line + 1, condition=condition)
        insert_variable_usage(variable, sourcefile, first_use_line + 2, condition=condition)
    else:
        warnings.warn("Variable %s not found %s in routine %s" % (
            variable.name, variable.routine.module.name, variable.routine.name))


def add_after_call_routines(source_file, vault, counter):
    from .Classes import Current_block
    try:
        current_block = Current_block(vault[source_file.filename.replace(".f90", "")].main())
    except AttributeError:
        warnings.warn("No module %s" % source_file.filename)
        return

    for _index, _line in enumerate(source_file.lines):
        # Find routine calls in line
        routine_calls = calls_to_routines(_line)
        variables_to_truncate = []
        if not routine_calls:
            # If not, look if the code is entering or exiting a routine block
            current_block = in_which_block(_line, current_block, vault=vault)
        else:
            # Keep the condition if exists
            if _line.lower().strip().find("if(") == 0:
                condition = close_brackets(_line)

            else:
                condition = None
            # If there's some call look at each one of them
            for call in routine_calls:
                is_external = False
                called_routine = find_called_routine(call)
                # If the routine is apply_truncation, continue
                if called_routine == "apply_truncation":
                    continue
                called_arguments = find_call_arguments(call)
                if not called_arguments:
                    # If there's no argument, continue
                    continue

                # Check that we properly found the routine in the database
                from .Classes import Routine, Interface
                try:
                    # Find the procedure called (object)
                    procedure = vault.get_procedure(called_routine)
                    if isinstance(procedure, Routine):
                        pass
                    elif isinstance(procedure, Interface):
                        procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                                  routine=current_block.routine, vault=vault)
                except ProcedureNotFound:
                    warnings.warn("Routine not found in vault database %s" % called_routine)
                    is_external = True

                # If the routine is in the database
                if not is_external:
                    # Check each argument
                    for argument_index, argument in enumerate(called_arguments):
                        if argument.count("tmp_var_"):
                            # Skipping temporal variables
                            continue
                        # Look if argument is keyword argument
                        if is_keyword_argument(argument):
                            # In case it is a keyword argument, get the dummy argument using the name.
                            keyword = argument.split("=")[0].strip()
                            dummy_argument = procedure.get_variable_by_name(keyword)
                            if not dummy_argument:
                                raise AssertionError("Keyword variable not found: %s" % keyword)

                            if dummy_argument.type != "rpe_var":
                                # If the argument is not an rpe_var because it's passing the %val element of the struct
                                # it should be truncated anyway.
                                if dummy_argument.type in real_types and argument.count("%val"):
                                    variables_to_truncate.append(argument.replace("%val", "").strip())
                            elif dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                variables_to_truncate.append(argument)
                        else:
                            dummy_argument = procedure.get_variable_by_position(argument_index)
                            if not dummy_argument:
                                raise AssertionError("Variable not found in position: %i" % argument_index)
                            if dummy_argument.type != "rpe_var":
                                if dummy_argument.type == real_types and argument.count("%val"):
                                    # If the argument contains whichever operations, for sure is intent in
                                    is_c2f = is_call_to_function(argument)
                                    has_op = has_operations(argument)
                                    if not is_c2f and not has_op:
                                        argument = argument.split("%val", 1)[0]
                                        if argument:
                                            variables_to_truncate.append(argument.strip())
                                        else:
                                            raise AssertionError
                            elif dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                variables_to_truncate.append(argument)
                else:
                    # If the routine is external
                    for argument_index, argument in enumerate(called_arguments):
                        # TODO: Probably should find a better solution, the code should not identify an array as a
                        # Todo: ... function even if the name coincides.
                        argument_type = type_of_contents(argument, routine=current_block.routine, vault=vault)
                        if argument_type == "rpe_var":
                            variables_to_truncate.append(argument)
                        elif argument_type in real_types and argument.count("%val"):
                            variables_to_truncate.append(argument.replace("%val", "").strip())
        i = _index
        for variable in variables_to_truncate:
            i = insert_in_line(variable, source_file, current_block.routine, counter, i, vault=vault,
                               condition=condition)


def insert_in_line(var, source_file, routine, counter, line_index, vault=None, condition=None):
    inner_index = line_index
    var_object = retrieve_variable(var, routine=routine, vault=vault)
    # Need to solve parameter issue here
    if var_object.is_parameter:
        return inner_index
    elif var_object.intent == "in":
        return inner_index

    if condition is None:
        condition = ""
    else:
        condition = condition

    sbits_assignation = "    %s%%sbits = emulator_variable_precisions(%i)"
    apply_truncation = "    CALL apply_truncation(%s) "

    if var_object.id is None:
        var_object.id = counter.count
        var_object.mutable = True
        counter.up()
    if var.count('='):
        var = var.split('=')[1].strip()

    source_file.lines.insert(inner_index + 1, condition + sbits_assignation % (var, var_object.id))
    source_file.lines.insert(inner_index + 2, condition + apply_truncation % var)

    return inner_index + 2


def replace_variable_exact_match(piece_to_replace, replacement, str):
    # In case our string does not start with a - or +, we'll add the \b to indicate a word border
    if piece_to_replace.strip()[0] not in ["-", "+"]:
        piece_to_replace = "\\b" + piece_to_replace.strip()
    # Adapt the string to fulfill regex rules
    piece_to_replace = piece_to_replace.replace("+", r"\+")
    piece_to_replace = piece_to_replace.replace("-", r"\-")
    piece_to_replace = piece_to_replace.replace("*", r"\*")
    if piece_to_replace.count("("):
        piece_to_replace = piece_to_replace.replace("(", "\s*\(")
        # Avoid matching structures
        if not piece_to_replace.count("%"):
            piece_to_replace = piece_to_replace.replace(")", "\)(?!%)")
        else:
            piece_to_replace = piece_to_replace.replace(")", "\)")
    else:
        piece_to_replace = piece_to_replace + "\\b(?!%val)"

    # Replace string using regex
    str = re.sub(piece_to_replace, replacement, str)
    return str


def temporal_variable_allocation_line(condition, temporal_variable_name, original_variable_name, dimension):
    if not dimension:
        assert isinstance(dimension, int)
        if dimension == 1:
            return condition + " allocate(%s( size(%s) ) )" % (temporal_variable_name, original_variable_name)
        elif dimension < 0:
            raise AssertionError("Dimension can not be negative!")
        else:
            addition = "size(%s,1)" % original_variable_name
            for i in range(2, dimension + 1):
                addition += ", size(%s,%i)" % (original_variable_name, i)
            return condition + " allocate(%s( %s ) )" % (temporal_variable_name, addition)
    else:
        return ""


def use_temporal_variable(sourcefile, line_index, argument, argument_index, routine=None, dummy=None, counter=None):
    tmp_var = "tmp_var_%i" % counter.count
    from .Classes import Variable
    var_object = Variable(routine, tmp_var, "rpe_var", dummy.dimension)
    routine.add_variable(var_object)
    counter.up()

    temporal_call = sourcefile.lines[line_index]
    if temporal_call.lower().strip().find("if(") == 0:
        condition = close_brackets(temporal_call)
    else:
        condition = ""

    # Temporal variable declaration
    temporal_variable_declaration = temporal_variable_allocation_line(condition, tmp_var, argument, dummy.dimension)

    # Allocation
    allocation = temporal_variable_allocation_line(condition, tmp_var, argument, dummy.dimension)

    # Deallocation
    if not dummy.dimension:
        deallocation = condition + "deallocate(%s)" % tmp_var
    else:
        deallocation = ""

    # Pre assignation
    preassignation = condition + "%s = %s" % (tmp_var, argument)

    # Post assignation
    postassignation = condition + "%s = %s" % (argument, tmp_var)

    # Insert declaration line
    declaration_line = line_index
    for i in reversed(range(line_index)):
        if sourcefile.lines[i].count("::"):
            declaration_line = i
            break
    sourcefile.lines.insert(declaration_line + 1, temporal_variable_declaration)

    line_index += 1
    # Insert allocation
    sourcefile.lines.insert(line_index, allocation)
    # Insert preassignation
    sourcefile.lines.insert(line_index + 1, preassignation)

    line_index += 2

    # Replace call
    call = replace_argument(temporal_call, argument_index, tmp_var)

    sourcefile.lines[line_index] = call

    sourcefile.lines.insert(line_index + 1, postassignation)

    # Insert deallocation
    sourcefile.lines.insert(line_index + 2, deallocation)


def replace_argument(string, _field, new_argument):
    # Keep the condition if exists
    if string.lower().strip().find("if(") == 0:
        condition = close_brackets(string)
        has_condition = True
        string = string.replace(condition, "%REPLACEMENT%")
    else:
        has_condition = False

    num_arguments, intervals = find_argument_intervals(string)
    interval = intervals[_field]
    start = interval[0]
    end = interval[1]
    first_segment = string[:start]
    last_segment = string[end:]
    string = "%s %s %s" % (first_segment, new_argument, last_segment)
    if has_condition:
        string = string.replace("%REPLACEMENT%", condition)
    return string
