import signal


class GracefulKiller:
    """
    This class is used to allow the user to stop an analysis at any point creating a pause_checkpoint.pkl which
    can be used in the future to restart the analysis from the same point.

    It uses the signal library to catch signals.
    """
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        self.filename = None

    def exit_gracefully(self, signum, frame):
        print("Exit has been activated, will write the checkpoint as soon as possible and exit.")
        self.kill_now = True


class PrecisionAnalysis:
    """
    The precision analysis class has the purpose of
    """
    def __init__(self,
                 # Few parameters needed to start a remote connection
                 remote_hostname=None,
                 remote_user=None,
                 remote_scratch=None,
                 max_running_jobs=10,
                 local_folder=None,

                 # Few experiment details
                 job_template=None,
                 reduced_precision_level=10,
                 experiment_name="PrecisionAnalysis",

                 # Few arguments that are required
                 vault=None,
                 accuracy_test=None,
                 analysis_variables=None,
                 forced_variables=None,

                 # In case of starting from a checkpoint
                 checkpoint_name=None,

                 ):
        from AutoRPE.UtilsWorkflow.Communicator import SSH

        # Saving all the arguments as elements of the class

        # Remote connection related parameters
        self.experiment_name = experiment_name
        self.remote_hostname = remote_hostname
        self.remote_user = remote_user
        self.remote_scratch = remote_scratch
        self.max_running_jobs = max_running_jobs
        self.local_folder = local_folder
        # Starting SSH connection
        self.communicator = SSH(user=self.remote_user, host=self.remote_hostname, remote_scratch=self.remote_scratch)

        # Parameters of the analysis
        self.reduced_precision_level = reduced_precision_level

        # Experiment specific related stuff
        self.accuracy_test = accuracy_test
        self.job_template = job_template
        self.vault = vault
        self.analysis_variables = analysis_variables
        if forced_variables is not None:
            self.forced_variables = forced_variables[:]
        else:
            self.forced_variables = None

        # In case of starting from a checkpoint
        self.checkpoint_name = checkpoint_name

        # Initialize an empty experiment driver
        self.experiment_driver = None

    def setup(self):
        """
        This method sets up the experiment driver in order to start with the experiment.
        """
        from AutoRPE.UtilsWorkflow.BinaryJob import BinarySearchJob
        from AutoRPE.UtilsWorkflow.Driver import ExperimentDriver
        from AutoRPE.UtilsWorkflow.CheckpointTools import recover_checkpoint

        vault = self.vault
        # Keeping the variables that have an id (it's precision can be changed in the code)
        variables = [v for v in vault.variables if v.id is not None]

        if self.analysis_variables:
            variables = [v for v in variables if v.id in self.analysis_variables]

        print("Starting analysis with %i variables" % len(variables))
        # Original set including all the variables
        original_set = [v.id for v in variables]

        # Create the starting job or recover it from a checkpoint
        if self.checkpoint_name is None:
            # Creating the first set that will include all the variables
            starting_job = BinarySearchJob(analysis_variables=original_set,
                                           forced_variables=self.forced_variables,
                                           reduced_precision_level=self.reduced_precision_level,
                                           communicator=self.communicator,
                                           vault=vault,
                                           template=self.job_template,
                                           accuracy_test=self.accuracy_test,
                                           local_folder=self.local_folder)
            # Defining it's level to 0
            starting_job.level = 0
            # Initializing an empty queue for the execution
            experiment = ExperimentDriver(starting_job)
        else:
            # Loading from checkpoint
            print("Resuming from %s" % self.checkpoint_name)
            experiment, starting_job, loops = recover_checkpoint(communicator=self.communicator, vault=vault,
                                                                 filename=self.checkpoint_name)
            # As the checkpoint can come from a different local platform, need to adjust template and local folder.
            for job in experiment.all():
                job.template = self.job_template
                job.local_folder = self.local_folder
            starting_job.vault = vault

        # Save the experiment driver in the class object
        self.experiment_driver = experiment

    def initial_check(self):
        from AutoRPE.UtilsWorkflow.BinaryJob import BinarySearchJob
        import time
        # Make a double-precision run to be sure that the accuracy test works as expected
        check_job = BinarySearchJob(
            analysis_variables=[],
            forced_variables=self.forced_variables,
            reduced_precision_level=self.reduced_precision_level,
            communicator=self.communicator,
            vault=self.vault,
            template=self.job_template,
            accuracy_test=self.accuracy_test,
            local_folder=self.local_folder
        )
        # Submit job and wait until it finishes
        check_job.submit()
        while check_job.remote_status != "COMPLETED" and check_job.remote_status != "FAILED":
            time.sleep(5)

        # Check that the job ran successfully
        if check_job.evaluate_success():
            print("Basic test was successful, continuing with the analysis")
        else:
            raise AssertionError("Basic simulation to check that the accuracy test works properly didn't pass")

    def run(self):
        ###############################################
        from AutoRPE.UtilsRPE.Functions import save_variable_list
        from AutoRPE.EvaluationScripts.result_explorer import create_explorer
        from AutoRPE.UtilsWorkflow.CheckpointTools import checkpoint
        import time

        # Before starting running the analysis, it is convenient to perform an initial check to assert that
        # setting all the analysis variables in double precision it passes the test.
        # Otherwise it would mean that we are trying to do something impossible.
        self.initial_check()

        # After that we will initialize the experiment driver
        self.setup()
        experiment = self.experiment_driver

        # Define a condition to keep iterating with the refinement
        def condition(_job):
            if _job.status == "FAILED":
                raise AssertionError("Whole set job ended with a failed result. Should check what happened.")
            # The condition of running will be True until the status of the job becomes SUCCESS
            return _job.status != "SUCCESS"

        # Start dynamic workflow
        try:
            change_counter = 0

            # Boolean to control weather we are waiting a minimum amount of time for each step or not.
            wait = True
            waiting_time = 10.
            # Use a signal catcher to handle experiment pause
            killer = GracefulKiller()
            while condition(experiment.seed_job):
                # Trying to avoid too much ssh polling by waiting if the loop time is less than a threshold
                loop_start_time = time.time()
                # Main loop
                # Sorting the queue to launch first the lower levels (bigger groups).
                experiment.sort(lower_first=True)
                # Launch one step (will loop through the jobs in the queue and take actions according to their status)
                step_change = experiment.step()

                # In case the status of any of the jobs from the experiment changed, print the status.
                if step_change:
                    experiment.print_status()
                # Check how many jobs did change status in order to save experiment checkpoints from time to time
                change_counter += step_change
                if change_counter > 10:
                    experiment.checkpoint()
                    change_counter = 0
                loop_end_time = time.time()

                # In case there hasn't been any change, this prevents the script of polling slurm too much
                if step_change == 0 and wait:
                    while loop_end_time - loop_start_time < waiting_time:
                        time.sleep(1)
                        loop_end_time = time.time()
                # If there was a signal during this loop:
                if killer.kill_now:
                    experiment.checkpoint(filename="paused_checkpoint.pkl")
                    print("Exiting...")
                    exit(0)
        except AssertionError as err:
            print(err)
            # For now skip the explorer part
            # Create HTML explorer to make result exploration easier
            # create_explorer(experiment=experiment, vault=vault)
            return False

        # Save lists of variables
        save_variable_list(experiment.seed_job.variable_set,
                           "%s_final_set.txt" % self.experiment_name,
                           vault=self.vault)
        save_variable_list(experiment.seed_job.banned_variables,
                           "%s_banned_variables.txt" % self.experiment_name,
                           vault=self.vault)

        # Saving a last checkpoint
        checkpoint(experiment, 1000, filename="end_checkpoint.pkl")

        # Probably not working right now
        # create_explorer(experiment=experiment, vault=self.vault)
        # Return success
        return True
